@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright © 2018-2019 Damir Jelić <poljar@termina.org.uk>
 #
 # Permission to use, copy, modify, and/or distribute this software for
@@ -16,7 +14,7 @@
 
 """nio to-device events.
 
-To-device events are events that are sent dirrectly between two devices instead
+To-device events are events that are sent directly between two devices instead
 of normally sending events in a room.
 
 To-device events can be sent to a specific device of a user or to all devices
@@ -24,13 +22,22 @@ of a user.
 
 """
 
-from typing import Any, Dict, List, Optional, Union
-from copy import deepcopy
+from __future__ import annotations
 
+from copy import deepcopy
 from dataclasses import dataclass, field
+from typing import Any, Dict, Optional, Union
 
 from ..schemas import Schemas
-from .misc import BadEventType, verify, logger
+from .common import (
+    KeyVerificationAcceptMixin,
+    KeyVerificationCancelMixin,
+    KeyVerificationEventMixin,
+    KeyVerificationKeyMixin,
+    KeyVerificationMacMixin,
+    KeyVerificationStartMixin,
+)
+from .misc import BadEventType, logger, verify
 
 
 @dataclass
@@ -50,8 +57,9 @@ class ToDeviceEvent:
 
     @classmethod
     @verify(Schemas.to_device)
-    def parse_event(cls, event_dict):
-        # type: (Dict) -> Optional[Union[ToDeviceEvent, BadEventType]]
+    def parse_event(
+        cls, event_dict: Dict
+    ) -> Optional[Union[ToDeviceEvent, BadEventType]]:
         """Parse a to-device event and create a higher level event object.
 
         This function parses the type of the to-device event and produces a
@@ -84,7 +92,7 @@ class ToDeviceEvent:
         elif event_dict["type"] == "m.room_key_request":
             return BaseRoomKeyRequest.parse_event(event_dict)
 
-        return None
+        return UnknownToDeviceEvent.from_dict(event_dict)
 
     @classmethod
     @verify(Schemas.room_encrypted)
@@ -109,8 +117,9 @@ class ToDeviceEvent:
         if content["algorithm"] == "m.olm.v1.curve25519-aes-sha2":
             return OlmEvent.from_dict(event_dict)
 
-        logger.warn("Received an encrypted event with an unknown "
-                    "algorithm {}.".format(content["algorithm"]))
+        logger.warning(
+            f"Received an encrypted event with an unknown algorithm {content['algorithm']}."
+        )
 
         return None
 
@@ -122,16 +131,17 @@ class ToDeviceEvent:
             parsed_dict (dict): The dictionary representation of the event.
 
         """
-        raise NotImplementedError()
+        raise NotImplementedError
 
 
 @dataclass
 class BaseRoomKeyRequest(ToDeviceEvent):
     """Base class for room key requests.
-        requesting_device_id (str): The id of the device that is requesting the
-            key.
-        request_id (str): A unique identifier for the request.
+    requesting_device_id (str): The id of the device that is requesting the
+        key.
+    request_id (str): A unique identifier for the request.
     """
+
     requesting_device_id: str = field()
     request_id: str = field()
 
@@ -201,7 +211,7 @@ class RoomKeyRequestCancellation(BaseRoomKeyRequest):
 
 
 @dataclass
-class KeyVerificationEvent(ToDeviceEvent):
+class KeyVerificationEvent(KeyVerificationEventMixin, ToDeviceEvent):
     """Base class for key verification events.
 
     Attributes:
@@ -210,11 +220,9 @@ class KeyVerificationEvent(ToDeviceEvent):
 
     """
 
-    transaction_id: str = field()
-
 
 @dataclass
-class KeyVerificationStart(KeyVerificationEvent):
+class KeyVerificationStart(KeyVerificationStartMixin, KeyVerificationEvent):
     """Event signaling the start of a SAS key verification process.
 
     Attributes:
@@ -232,13 +240,6 @@ class KeyVerificationStart(KeyVerificationEvent):
 
     """
 
-    from_device: str = field()
-    method: str = field()
-    key_agreement_protocols: List[str] = field()
-    hashes: List[str] = field()
-    message_authentication_codes: List[str] = field()
-    short_authentication_string: List[str] = field()
-
     @classmethod
     @verify(Schemas.key_verification_start)
     def from_dict(cls, parsed_dict):
@@ -257,7 +258,7 @@ class KeyVerificationStart(KeyVerificationEvent):
 
 
 @dataclass
-class KeyVerificationAccept(KeyVerificationEvent):
+class KeyVerificationAccept(KeyVerificationAcceptMixin, KeyVerificationEvent):
     """Event signaling that the SAS verification start has been accepted.
 
     Attributes:
@@ -273,12 +274,6 @@ class KeyVerificationAccept(KeyVerificationEvent):
 
     """
 
-    commitment: str = field()
-    key_agreement_protocol: str = field()
-    hash: str = field()
-    message_authentication_code: str = field()
-    short_authentication_string: List[str] = field()
-
     @classmethod
     @verify(Schemas.key_verification_accept)
     def from_dict(cls, parsed_dict):
@@ -296,7 +291,7 @@ class KeyVerificationAccept(KeyVerificationEvent):
 
 
 @dataclass
-class KeyVerificationKey(KeyVerificationEvent):
+class KeyVerificationKey(KeyVerificationKeyMixin, KeyVerificationEvent):
     """Event carrying a key verification key.
 
     After this event is received the short authentication string can be shown
@@ -308,8 +303,6 @@ class KeyVerificationKey(KeyVerificationEvent):
 
     """
 
-    key: str = field()
-
     @classmethod
     @verify(Schemas.key_verification_key)
     def from_dict(cls, parsed_dict):
@@ -323,7 +316,7 @@ class KeyVerificationKey(KeyVerificationEvent):
 
 
 @dataclass
-class KeyVerificationMac(KeyVerificationEvent):
+class KeyVerificationMac(KeyVerificationMacMixin, KeyVerificationEvent):
     """Event holding a message authentication code of the verification process.
 
     After this event is received the device that we are verifying will be
@@ -339,9 +332,6 @@ class KeyVerificationMac(KeyVerificationEvent):
 
     """
 
-    mac: Dict[str, str] = field()
-    keys: str = field()
-
     @classmethod
     @verify(Schemas.key_verification_mac)
     def from_dict(cls, parsed_dict):
@@ -356,7 +346,7 @@ class KeyVerificationMac(KeyVerificationEvent):
 
 
 @dataclass
-class KeyVerificationCancel(KeyVerificationEvent):
+class KeyVerificationCancel(KeyVerificationCancelMixin, KeyVerificationEvent):
     """Event signaling that a key verification process has been canceled.
 
     Attributes:
@@ -366,9 +356,6 @@ class KeyVerificationCancel(KeyVerificationEvent):
 
     """
 
-    code: str = field()
-    reason: str = field()
-
     @classmethod
     @verify(Schemas.key_verification_cancel)
     def from_dict(cls, parsed_dict):
@@ -422,17 +409,14 @@ class OlmEvent(EncryptedToDeviceEvent):
         ciphertext = content["ciphertext"]
         sender_key = content["sender_key"]
 
-        tx_id = (event_dict["unsigned"].get("transaction_id", None)
-                 if "unsigned" in event_dict else None)
-
-        return cls(
-            event_dict,
-            event_dict["sender"],
-            sender_key,
-            ciphertext,
-            tx_id
+        tx_id = (
+            event_dict["unsigned"].get("transaction_id", None)
+            if "unsigned" in event_dict
+            else None
         )
 
+        return cls(event_dict, event_dict["sender"], sender_key, ciphertext, tx_id)
+
 
 @dataclass
 class DummyEvent(ToDeviceEvent):
@@ -453,12 +437,7 @@ class DummyEvent(ToDeviceEvent):
     @classmethod
     @verify(Schemas.dummy_event)
     def from_dict(cls, event_dict, sender, sender_key):
-        return cls(
-            event_dict,
-            sender,
-            sender_key,
-            event_dict["sender_device"]
-        )
+        return cls(event_dict, sender, sender_key, event_dict["sender_device"])
 
 
 @dataclass
@@ -495,7 +474,7 @@ class RoomKeyEvent(ToDeviceEvent):
             sender_key,
             content["room_id"],
             content["session_id"],
-            content["algorithm"]
+            content["algorithm"],
         )
 
 
@@ -533,5 +512,29 @@ class ForwardedRoomKeyEvent(RoomKeyEvent):
             sender_key,
             content["room_id"],
             content["session_id"],
-            content["algorithm"]
+            content["algorithm"],
+        )
+
+
+@dataclass
+class UnknownToDeviceEvent(ToDeviceEvent):
+    """A ToDeviceEvent which we do not understand.
+
+    This event is created every time nio tries to parse an event of an unknown
+    type. Since custom and extensible events are a feature of Matrix this
+    allows clients to use custom events but care should be taken that the
+    clients will be responsible to validate and type check the event.
+
+    Attributes:
+        type (str): The type of the event.
+    """
+
+    type: str = field()
+
+    @classmethod
+    def from_dict(cls, event_dict):
+        return cls(
+            event_dict,
+            event_dict["sender"],
+            event_dict["type"],
         )
